@@ -99,6 +99,12 @@ class ExternalTraceServer(tsi.TraceServerInterface):
     ) -> tsi.EnsureProjectExistsRes:
         return self._internal_trace_server.ensure_project_exists(entity, project)
 
+    def otel_export(self, req: tsi.OtelExportReq) -> tsi.OtelExportRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        if req.wb_user_id is not None:
+            req.wb_user_id = self._idc.ext_to_int_user_id(req.wb_user_id)
+        return self._ref_apply(self._internal_trace_server.otel_export, req)
+
     def call_start(self, req: tsi.CallStartReq) -> tsi.CallStartRes:
         req.start.project_id = self._idc.ext_to_int_project_id(req.start.project_id)
         if req.start.wb_run_id is not None:
@@ -236,6 +242,8 @@ class ExternalTraceServer(tsi.TraceServerInterface):
 
     def obj_create(self, req: tsi.ObjCreateReq) -> tsi.ObjCreateRes:
         req.obj.project_id = self._idc.ext_to_int_project_id(req.obj.project_id)
+        if req.obj.wb_user_id is not None:
+            req.obj.wb_user_id = self._idc.ext_to_int_user_id(req.obj.wb_user_id)
         return self._ref_apply(self._internal_trace_server.obj_create, req)
 
     def obj_read(self, req: tsi.ObjReadReq) -> tsi.ObjReadRes:
@@ -257,6 +265,10 @@ class ExternalTraceServer(tsi.TraceServerInterface):
             obj.project_id = original_project_id
         return res
 
+    def obj_delete(self, req: tsi.ObjDeleteReq) -> tsi.ObjDeleteRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        return self._ref_apply(self._internal_trace_server.obj_delete, req)
+
     def table_create(self, req: tsi.TableCreateReq) -> tsi.TableCreateRes:
         req.table.project_id = self._idc.ext_to_int_project_id(req.table.project_id)
         return self._ref_apply(self._internal_trace_server.table_create, req)
@@ -265,6 +277,14 @@ class ExternalTraceServer(tsi.TraceServerInterface):
         req.project_id = self._idc.ext_to_int_project_id(req.project_id)
         return self._ref_apply(self._internal_trace_server.table_update, req)
 
+    def table_create_from_digests(
+        self, req: tsi.TableCreateFromDigestsReq
+    ) -> tsi.TableCreateFromDigestsRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        return self._ref_apply(
+            self._internal_trace_server.table_create_from_digests, req
+        )
+
     def table_query(self, req: tsi.TableQueryReq) -> tsi.TableQueryRes:
         req.project_id = self._idc.ext_to_int_project_id(req.project_id)
         return self._ref_apply(self._internal_trace_server.table_query, req)
@@ -277,10 +297,17 @@ class ExternalTraceServer(tsi.TraceServerInterface):
             self._internal_trace_server.table_query_stream, req
         )
 
+    # This is a legacy endpoint, it should be removed once the client is mostly updated
     def table_query_stats(self, req: tsi.TableQueryStatsReq) -> tsi.TableQueryStatsRes:
         req.project_id = self._idc.ext_to_int_project_id(req.project_id)
         return self._ref_apply(self._internal_trace_server.table_query_stats, req)
 
+    def table_query_stats_batch(
+        self, req: tsi.TableQueryStatsBatchReq
+    ) -> tsi.TableQueryStatsBatchRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        return self._ref_apply(self._internal_trace_server.table_query_stats_batch, req)
+
     def refs_read_batch(self, req: tsi.RefsReadBatchReq) -> tsi.RefsReadBatchRes:
         return self._ref_apply(self._internal_trace_server.refs_read_batch, req)
 
@@ -294,6 +321,10 @@ class ExternalTraceServer(tsi.TraceServerInterface):
         # Special case where refs can never be part of the request
         return self._internal_trace_server.file_content_read(req)
 
+    def files_stats(self, req: tsi.FilesStatsReq) -> tsi.FilesStatsRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        return self._ref_apply(self._internal_trace_server.files_stats, req)
+
     def feedback_create(self, req: tsi.FeedbackCreateReq) -> tsi.FeedbackCreateRes:
         req.project_id = self._idc.ext_to_int_project_id(req.project_id)
         original_user_id = req.wb_user_id
@@ -306,6 +337,20 @@ class ExternalTraceServer(tsi.TraceServerInterface):
         res.wb_user_id = original_user_id
         return res
 
+    def feedback_create_batch(
+        self, req: tsi.FeedbackCreateBatchReq
+    ) -> tsi.FeedbackCreateBatchRes:
+        for feedback_req in req.batch:
+            feedback_req.project_id = self._idc.ext_to_int_project_id(
+                feedback_req.project_id
+            )
+            if feedback_req.wb_user_id is not None:
+                feedback_req.wb_user_id = self._idc.ext_to_int_user_id(
+                    feedback_req.wb_user_id
+                )
+        res = self._ref_apply(self._internal_trace_server.feedback_create_batch, req)
+        return res
+
     def feedback_query(self, req: tsi.FeedbackQueryReq) -> tsi.FeedbackQueryRes:
         original_project_id = req.project_id
         req.project_id = self._idc.ext_to_int_project_id(original_project_id)
@@ -316,11 +361,10 @@ class ExternalTraceServer(tsi.TraceServerInterface):
                 if feedback["project_id"] != req.project_id:
                     raise ValueError("Internal Error - Project Mismatch")
                 feedback["project_id"] = original_project_id
-            if "wb_user_id" in feedback:
-                if feedback["wb_user_id"] is not None:
-                    feedback["wb_user_id"] = self._idc.int_to_ext_user_id(
-                        feedback["wb_user_id"]
-                    )
+            if "wb_user_id" in feedback and feedback["wb_user_id"] is not None:
+                feedback["wb_user_id"] = self._idc.int_to_ext_user_id(
+                    feedback["wb_user_id"]
+                )
         return res
 
     def feedback_purge(self, req: tsi.FeedbackPurgeReq) -> tsi.FeedbackPurgeRes:
@@ -376,3 +420,36 @@ class ExternalTraceServer(tsi.TraceServerInterface):
         req.project_id = self._idc.ext_to_int_project_id(req.project_id)
         res = self._ref_apply(self._internal_trace_server.completions_create, req)
         return res
+
+    # Streaming completions – simply proxy through after converting project ID.
+    def completions_create_stream(
+        self, req: tsi.CompletionsCreateReq
+    ) -> typing.Iterator[dict[str, typing.Any]]:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        # The streamed chunks contain no project-scoped references, so we can
+        # forward directly without additional ref conversion.
+        return self._internal_trace_server.completions_create_stream(req)
+
+    def project_stats(self, req: tsi.ProjectStatsReq) -> tsi.ProjectStatsRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        return self._ref_apply(self._internal_trace_server.project_stats, req)
+
+    def threads_query_stream(
+        self, req: tsi.ThreadsQueryReq
+    ) -> Iterator[tsi.ThreadSchema]:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        return self._stream_ref_apply(
+            self._internal_trace_server.threads_query_stream, req
+        )
+
+    def evaluate_model(self, req: tsi.EvaluateModelReq) -> tsi.EvaluateModelRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        if req.wb_user_id is not None:
+            req.wb_user_id = self._idc.ext_to_int_user_id(req.wb_user_id)
+        return self._ref_apply(self._internal_trace_server.evaluate_model, req)
+
+    def evaluation_status(
+        self, req: tsi.EvaluationStatusReq
+    ) -> tsi.EvaluationStatusRes:
+        req.project_id = self._idc.ext_to_int_project_id(req.project_id)
+        return self._ref_apply(self._internal_trace_server.evaluation_status, req)
