@@ -38,3 +38,15 @@ def feedback_type_is_annotation(feedback_type: str) -> bool:
 
 def feedback_type_is_runnable(feedback_type: str) -> bool:
     return feedback_type.startswith(RUNNABLE_FEEDBACK_TYPE_PREFIX)
+
+
+def runnable_feedback_selector(name: str) -> str:
+    return f"feedback.[{RUNNABLE_FEEDBACK_TYPE_PREFIX}.{name}]"
+
+
+def runnable_feedback_output_selector(name: str) -> str:
+    return f"{runnable_feedback_selector(name)}.payload.output"
+
+
+def runnable_feedback_runnable_ref_selector(name: str) -> str:
+    return f"{runnable_feedback_selector(name)}.runnable_ref"
