@@ -3,7 +3,12 @@ from collections.abc import Iterator
 from enum import Enum
 from typing import Any, Literal, Optional, Protocol, Union
 
-from pydantic import BaseModel, ConfigDict, Field, field_serializer
+from pydantic import (
+    BaseModel,
+    ConfigDict,
+    Field,
+    field_serializer,
+)
 from typing_extensions import TypedDict
 
 from weave.trace_server.interface.query import Query
@@ -13,6 +18,12 @@ WB_USER_ID_DESCRIPTION = (
 )
 
 
+class BaseModelStrict(BaseModel):
+    """Base model with strict validation that forbids extra fields."""
+
+    model_config = ConfigDict(extra="forbid")
+
+
 class ExtraKeysTypedDict(TypedDict):
     pass
 
@@ -59,6 +70,7 @@ class TraceStatus(str, Enum):
     SUCCESS = "success"
     ERROR = "error"
     RUNNING = "running"
+    DESCENDANT_ERROR = "descendant_error"
 
 
 class WeaveSummarySchema(ExtraKeysTypedDict, total=False):
@@ -72,6 +84,7 @@ class WeaveSummarySchema(ExtraKeysTypedDict, total=False):
 
 class SummaryInsertMap(ExtraKeysTypedDict, total=False):
     usage: dict[str, LLMUsageSchema]
+    status_counts: dict[TraceStatus, int]
 
 
 class SummaryMap(SummaryInsertMap, total=False):
@@ -91,6 +104,10 @@ class CallSchema(BaseModel):
     trace_id: str
     # Parent ID is optional because the call may be a root
     parent_id: Optional[str] = None
+    # Thread ID is optional
+    thread_id: Optional[str] = None
+    # Turn ID is optional
+    turn_id: Optional[str] = None
 
     # Start time is required
     started_at: datetime.datetime
@@ -115,9 +132,16 @@ class CallSchema(BaseModel):
     # WB Metadata
     wb_user_id: Optional[str] = None
     wb_run_id: Optional[str] = None
+    wb_run_step: Optional[int] = None
 
     deleted_at: Optional[datetime.datetime] = None
 
+    # Size of metadata storage for this call
+    storage_size_bytes: Optional[int] = None
+
+    # Total size of metadata storage for the entire trace
+    total_storage_size_bytes: Optional[int] = None
+
     @field_serializer("attributes", "summary", when_used="unless-none")
     def serialize_typed_dicts(self, v: dict[str, Any]) -> dict[str, Any]:
         return dict(v)
@@ -139,6 +163,10 @@ class StartedCallSchemaForInsert(BaseModel):
     trace_id: Optional[str] = None  # Will be generated if not provided
     # Parent ID is optional because the call may be a root
     parent_id: Optional[str] = None
+    # Thread ID is optional
+    thread_id: Optional[str] = None
+    # Turn ID is optional
+    turn_id: Optional[str] = None
 
     # Start time is required
     started_at: datetime.datetime
@@ -151,6 +179,7 @@ class StartedCallSchemaForInsert(BaseModel):
     # WB Metadata
     wb_user_id: Optional[str] = Field(None, description=WB_USER_ID_DESCRIPTION)
     wb_run_id: Optional[str] = None
+    wb_run_step: Optional[int] = None
 
 
 class EndedCallSchemaForInsert(BaseModel):
@@ -184,8 +213,12 @@ class ObjSchema(BaseModel):
     is_latest: int
     kind: str
     base_object_class: Optional[str]
+    leaf_object_class: Optional[str] = None
     val: Any
 
+    wb_user_id: Optional[str] = Field(None, description=WB_USER_ID_DESCRIPTION)
+    size_bytes: Optional[int] = None
+
 
 class ObjSchemaForInsert(BaseModel):
     project_id: str
@@ -194,9 +227,11 @@ class ObjSchemaForInsert(BaseModel):
     builtin_object_class: Optional[str] = None
     # Keeping `set_base_object_class` here until it is successfully removed from UI client
     set_base_object_class: Optional[str] = Field(
-        include=False, default=None, deprecated=True
+        exclude=True, default=None, deprecated=True
     )
 
+    wb_user_id: Optional[str] = Field(None, description=WB_USER_ID_DESCRIPTION)
+
     def model_post_init(self, __context: Any) -> None:
         # If set_base_object_class is provided, use it to set builtin_object_class for backwards compatibility
         if self.set_base_object_class is not None and self.builtin_object_class is None:
@@ -208,7 +243,29 @@ class TableSchemaForInsert(BaseModel):
     rows: list[dict[str, Any]]
 
 
-class CallStartReq(BaseModel):
+class OtelExportReq(BaseModel):
+    model_config = ConfigDict(arbitrary_types_allowed=True)
+    project_id: str
+    # traces must be ExportTraceServiceRequest payload but allowing Any removes the proto package as a requirement.
+    traces: Any
+    wb_user_id: Optional[str] = Field(None, description=WB_USER_ID_DESCRIPTION)
+
+
+class ExportTracePartialSuccess(BaseModel):
+    rejected_spans: int
+    error_message: str
+
+
+# Spec requires that the response be of type Export<signal>ServiceResponse
+# https://opentelemetry.io/docs/specs/otlp/
+class OtelExportRes(BaseModel):
+    partial_success: Optional[ExportTracePartialSuccess] = Field(
+        default=None,
+        description="The details of a partially successful export request. When None or rejected_spans is 0, the request was fully accepted.",
+    )
+
+
+class CallStartReq(BaseModelStrict):
     start: StartedCallSchemaForInsert
 
 
@@ -217,7 +274,7 @@ class CallStartRes(BaseModel):
     trace_id: str
 
 
-class CallEndReq(BaseModel):
+class CallEndReq(BaseModelStrict):
     end: EndedCallSchemaForInsert
 
 
@@ -225,17 +282,37 @@ class CallEndRes(BaseModel):
     pass
 
 
-class CallReadReq(BaseModel):
+class CallBatchStartMode(BaseModel):
+    mode: str = "start"
+    req: CallStartReq
+
+
+class CallBatchEndMode(BaseModel):
+    mode: str = "end"
+    req: CallEndReq
+
+
+class CallCreateBatchReq(BaseModelStrict):
+    batch: list[Union[CallBatchStartMode, CallBatchEndMode]]
+
+
+class CallCreateBatchRes(BaseModel):
+    res: list[Union[CallStartRes, CallEndRes]]
+
+
+class CallReadReq(BaseModelStrict):
     project_id: str
     id: str
     include_costs: Optional[bool] = False
+    include_storage_size: Optional[bool] = False
+    include_total_storage_size: Optional[bool] = False
 
 
 class CallReadRes(BaseModel):
     call: Optional[CallSchema]
 
 
-class CallsDeleteReq(BaseModel):
+class CallsDeleteReq(BaseModelStrict):
     project_id: str
     call_ids: list[str]
 
@@ -260,6 +337,7 @@ class CompletionsCreateRequestInputs(BaseModel):
     modalities: Optional[list] = None
     presence_penalty: Optional[float] = None
     frequency_penalty: Optional[float] = None
+    stream: Optional[bool] = None
     logit_bias: Optional[dict] = None
     user: Optional[str] = None
     # openai v1.0+ new params
@@ -277,7 +355,7 @@ class CompletionsCreateRequestInputs(BaseModel):
     api_version: Optional[str] = None
 
 
-class CompletionsCreateReq(BaseModel):
+class CompletionsCreateReq(BaseModelStrict):
     project_id: str
     inputs: CompletionsCreateRequestInputs
     wb_user_id: Optional[str] = Field(None, description=WB_USER_ID_DESCRIPTION)
@@ -291,19 +369,21 @@ class CompletionsCreateRes(BaseModel):
     weave_call_id: Optional[str] = None
 
 
-class CallsFilter(BaseModel):
+class CallsFilter(BaseModelStrict):
     op_names: Optional[list[str]] = None
     input_refs: Optional[list[str]] = None
     output_refs: Optional[list[str]] = None
     parent_ids: Optional[list[str]] = None
     trace_ids: Optional[list[str]] = None
     call_ids: Optional[list[str]] = None
+    thread_ids: Optional[list[str]] = None
+    turn_ids: Optional[list[str]] = None
     trace_roots_only: Optional[bool] = None
     wb_user_ids: Optional[list[str]] = None
     wb_run_ids: Optional[list[str]] = None
 
 
-class SortBy(BaseModel):
+class SortBy(BaseModelStrict):
     # Field should be a key of `CallSchema`. For dictionary fields
     # (`attributes`, `inputs`, `outputs`, `summary`), the field can be
     # dot-separated.
@@ -312,7 +392,7 @@ class SortBy(BaseModel):
     direction: Literal["asc", "desc"]
 
 
-class CallsQueryReq(BaseModel):
+class CallsQueryReq(BaseModelStrict):
     project_id: str
     filter: Optional[CallsFilter] = None
     limit: Optional[int] = None
@@ -330,33 +410,74 @@ class CallsQueryReq(BaseModel):
         description="Beta, subject to change. If true, the response will"
         " include feedback for each call.",
     )
+    include_storage_size: Optional[bool] = Field(
+        default=False,
+        description="Beta, subject to change. If true, the response will"
+        " include the storage size for a call.",
+    )
+    include_total_storage_size: Optional[bool] = Field(
+        default=False,
+        description="Beta, subject to change. If true, the response will"
+        " include the total storage size for a trace.",
+    )
 
     # TODO: type this with call schema columns, following the same rules as
     # SortBy and thus GetFieldOperator.get_field_ (without direction)
     columns: Optional[list[str]] = None
-    # columns to expand, i.e. refs to other objects, can be nested
+
+    # Columns to expand, i.e. refs to other objects, can be nested
+    # Also used to provide a list of refs to expand when filtering or sorting.
+    # Requests to filter or order calls by sub fields in columns that have
+    # refs in their path must provide paths to all refs in the expand_columns.
+    # When filtering and ordering, expand_columns can include paths to objects
+    # that are stored in the table_rows table.
+    # TODO: support expand_columns for refs to objects in table_rows (dataset rows)
     expand_columns: Optional[list[str]] = Field(
         default=None,
         examples=[["inputs.self.message", "inputs.model.prompt"]],
         description="Columns to expand, i.e. refs to other objects",
     )
+    # Controls whether or not to return expanded ref columns. In most clients,
+    # refs are resolved recursively by making additional api calls, either for
+    # performance or convenience reasons. In that case, we do not want to return
+    # resolved refs. However, expand_columns still must contain paths to all
+    # refs when filtering or sorting. Set this value to false to filter/order
+    # by refs but rely on client methods for actually resolving the values. The
+    # default is to resolve and return expanded values when expand_columns is set.
+    return_expanded_column_values: Optional[bool] = Field(
+        default=True,
+        description="If true, the response will include raw values for expanded columns. "
+        "If false, the response expand_columns will only be used for filtering and ordering. "
+        "This is useful for clients that want to resolve refs themselves, e.g. for performance reasons.",
+    )
 
 
 class CallsQueryRes(BaseModel):
     calls: list[CallSchema]
 
 
-class CallsQueryStatsReq(BaseModel):
+class CallsQueryStatsReq(BaseModelStrict):
     project_id: str
     filter: Optional[CallsFilter] = None
     query: Optional[Query] = None
+    limit: Optional[int] = None
+    include_total_storage_size: Optional[bool] = False
+    # List of columns that include refs to objects or table rows that require
+    # expansion during filtering or ordering. Required when filtering
+    # on reffed fields.
+    expand_columns: Optional[list[str]] = Field(
+        default=None,
+        examples=[["inputs.self.message", "inputs.model.prompt"]],
+        description="Columns with refs to objects or table rows that require expansion during filtering or ordering.",
+    )
 
 
 class CallsQueryStatsRes(BaseModel):
     count: int
+    total_storage_size_bytes: Optional[int] = None
 
 
-class CallUpdateReq(BaseModel):
+class CallUpdateReq(BaseModelStrict):
     # required for all updates
     project_id: str
     call_id: str
@@ -372,7 +493,7 @@ class CallUpdateRes(BaseModel):
     pass
 
 
-class OpCreateReq(BaseModel):
+class OpCreateReq(BaseModelStrict):
     op_obj: ObjSchemaForInsert
 
 
@@ -380,7 +501,7 @@ class OpCreateRes(BaseModel):
     digest: str
 
 
-class OpReadReq(BaseModel):
+class OpReadReq(BaseModelStrict):
     project_id: str
     name: str
     digest: str
@@ -395,7 +516,7 @@ class OpVersionFilter(BaseModel):
     latest_only: Optional[bool] = None
 
 
-class OpQueryReq(BaseModel):
+class OpQueryReq(BaseModelStrict):
     project_id: str
     filter: Optional[OpVersionFilter] = None
 
@@ -404,7 +525,7 @@ class OpQueryRes(BaseModel):
     op_objs: list[ObjSchema]
 
 
-class ObjCreateReq(BaseModel):
+class ObjCreateReq(BaseModelStrict):
     obj: ObjSchemaForInsert
 
 
@@ -412,22 +533,33 @@ class ObjCreateRes(BaseModel):
     digest: str  #
 
 
-class ObjReadReq(BaseModel):
+class ObjReadReq(BaseModelStrict):
     project_id: str
     object_id: str
     digest: str
 
+    metadata_only: Optional[bool] = Field(
+        default=False,
+        description="If true, the `val` column is not read from the database and is empty."
+        "All other fields are returned.",
+    )
+
 
 class ObjReadRes(BaseModel):
     obj: ObjSchema
 
 
-class ObjectVersionFilter(BaseModel):
+class ObjectVersionFilter(BaseModelStrict):
     base_object_classes: Optional[list[str]] = Field(
         default=None,
         description="Filter objects by their base classes",
         examples=[["Model"], ["Dataset"]],
     )
+    leaf_object_classes: Optional[list[str]] = Field(
+        default=None,
+        description="Filter objects by their leaf classes",
+        examples=[["Model"], ["Dataset"], ["LLMStructuredCompletionModel"]],
+    )
     object_ids: Optional[list[str]] = Field(
         default=None,
         description="Filter objects by their IDs",
@@ -445,7 +577,7 @@ class ObjectVersionFilter(BaseModel):
     )
 
 
-class ObjQueryReq(BaseModel):
+class ObjQueryReq(BaseModelStrict):
     project_id: str = Field(
         description="The ID of the project to query", examples=["user/project"]
     )
@@ -474,16 +606,42 @@ class ObjQueryReq(BaseModel):
         description="If true, the `val` column is not read from the database and is empty."
         "All other fields are returned.",
     )
+    include_storage_size: Optional[bool] = Field(
+        default=False,
+        description="If true, the `size_bytes` column is returned.",
+    )
+
+
+class ObjDeleteReq(BaseModelStrict):
+    project_id: str
+    object_id: str
+    digests: Optional[list[str]] = Field(
+        default=None,
+        description="List of digests to delete. If not provided, all digests for the object will be deleted.",
+    )
+
+
+class ObjDeleteRes(BaseModel):
+    num_deleted: int
 
 
 class ObjQueryRes(BaseModel):
     objs: list[ObjSchema]
 
 
-class TableCreateReq(BaseModel):
+class TableCreateReq(BaseModelStrict):
     table: TableSchemaForInsert
 
 
+class TableCreateFromDigestsReq(BaseModelStrict):
+    project_id: str
+    row_digests: list[str]
+
+
+class TableCreateFromDigestsRes(BaseModel):
+    digest: str
+
+
 """
 The `TableUpdateSpec` pattern is as follows, where `OPERATION` is globally unique. This
 follows a similar pattern as our `Query` definitions.
@@ -563,7 +721,7 @@ class TableInsertSpec(BaseModel):
 TableUpdateSpec = Union[TableAppendSpec, TablePopSpec, TableInsertSpec]
 
 
-class TableUpdateReq(BaseModel):
+class TableUpdateReq(BaseModelStrict):
     project_id: str
     base_digest: str
     updates: list[TableUpdateSpec]
@@ -577,7 +735,7 @@ class TableUpdateRes(BaseModel):
     # As a result, we might have servers in the wild that
     # do not support this field. Therefore, we want to ensure
     # that clients expecting this field will not break when
-    # they are targetting an older server. We should remove
+    # they are targeting an older server. We should remove
     # this default factory once we are sure that all servers
     # have been updated to support this field.
     updated_row_digests: list[str] = Field(
@@ -588,6 +746,7 @@ class TableUpdateRes(BaseModel):
 class TableRowSchema(BaseModel):
     digest: str
     val: Any
+    original_index: Optional[int] = None
 
 
 class TableCreateRes(BaseModel):
@@ -598,7 +757,7 @@ class TableCreateRes(BaseModel):
     # As a result, we might have servers in the wild that
     # do not support this field. Therefore, we want to ensure
     # that clients expecting this field will not break when
-    # they are targetting an older server. We should remove
+    # they are targeting an older server. We should remove
     # this default factory once we are sure that all servers
     # have been updated to support this field.
     row_digests: list[str] = Field(
@@ -606,7 +765,7 @@ class TableCreateRes(BaseModel):
     )
 
 
-class TableRowFilter(BaseModel):
+class TableRowFilter(BaseModelStrict):
     row_digests: Optional[list[str]] = Field(
         default=None,
         description="List of row digests to filter by",
@@ -619,7 +778,7 @@ class TableRowFilter(BaseModel):
     )
 
 
-class TableQueryReq(BaseModel):
+class TableQueryReq(BaseModelStrict):
     project_id: str = Field(
         description="The ID of the project", examples=["my_entity/my_project"]
     )
@@ -658,13 +817,31 @@ class TableQueryRes(BaseModel):
     rows: list[TableRowSchema]
 
 
-class TableQueryStatsReq(BaseModel):
+class TableQueryStatsReq(BaseModelStrict):
     project_id: str = Field(
         description="The ID of the project", examples=["my_entity/my_project"]
     )
     digest: str = Field(
         description="The digest of the table to query",
-        examples=["aonareimsvtl13apimtalpa4435rpmgnaemrpgmarltarstaorsnte134avrims"],
+    )
+
+
+class TableQueryStatsBatchReq(BaseModelStrict):
+    project_id: str = Field(
+        description="The ID of the project", examples=["my_entity/my_project"]
+    )
+
+    digests: Optional[list[str]] = Field(
+        description="The digests of the tables to query",
+        examples=[
+            "aonareimsvtl13apimtalpa4435rpmgnaemrpgmarltarstaorsnte134avrims",
+            "smirva431etnsroatsratlrampgrmeangmpr5344aplatmipa31ltvsmiераnoa",
+        ],
+        default=[],
+    )
+    include_storage_size: Optional[bool] = Field(
+        default=False,
+        description="If true, the `storage_size_bytes` column is returned.",
     )
 
 
@@ -672,7 +849,17 @@ class TableQueryStatsRes(BaseModel):
     count: int
 
 
-class RefsReadBatchReq(BaseModel):
+class TableStatsRow(BaseModel):
+    count: int
+    digest: str
+    storage_size_bytes: Optional[int] = None
+
+
+class TableQueryStatsBatchRes(BaseModel):
+    tables: list[TableStatsRow]
+
+
+class RefsReadBatchReq(BaseModelStrict):
     refs: list[str]
 
 
@@ -680,7 +867,12 @@ class RefsReadBatchRes(BaseModel):
     vals: list[Any]
 
 
-class FeedbackCreateReq(BaseModel):
+class FeedbackCreateReq(BaseModelStrict):
+    id: Optional[str] = Field(
+        default=None,
+        description="If provided by the client, this ID will be used for the feedback row instead of a server-generated one.",
+        examples=["018f1f2a-9c2b-7d3e-b5a1-8c9d2e4f6a7b"],
+    )
     project_id: str = Field(examples=["entity/project"])
     weave_ref: str = Field(examples=["weave:///entity/project/object/name:digest"])
     creator: Optional[str] = Field(default=None, examples=["Jane Smith"])
@@ -721,11 +913,12 @@ class FeedbackCreateRes(BaseModel):
 
 
 class Feedback(FeedbackCreateReq):
-    id: str
+    # Feedback is stricter than the create request, and must always have an id
+    id: str  # type: ignore[reportIncompatibleVariableOverride]
     created_at: datetime.datetime
 
 
-class FeedbackQueryReq(BaseModel):
+class FeedbackQueryReq(BaseModelStrict):
     project_id: str = Field(examples=["entity/project"])
     fields: Optional[list[str]] = Field(
         default=None, examples=[["id", "feedback_type", "payload.note"]]
@@ -743,7 +936,7 @@ class FeedbackQueryRes(BaseModel):
     result: list[dict[str, Any]]
 
 
-class FeedbackPurgeReq(BaseModel):
+class FeedbackPurgeReq(BaseModelStrict):
     project_id: str = Field(examples=["entity/project"])
     query: Query
 
@@ -760,7 +953,15 @@ class FeedbackReplaceRes(FeedbackCreateRes):
     pass
 
 
-class FileCreateReq(BaseModel):
+class FeedbackCreateBatchReq(BaseModelStrict):
+    batch: list[FeedbackCreateReq]
+
+
+class FeedbackCreateBatchRes(BaseModel):
+    res: list[FeedbackCreateRes]
+
+
+class FileCreateReq(BaseModelStrict):
     project_id: str
     name: str
     content: bytes
@@ -770,20 +971,28 @@ class FileCreateRes(BaseModel):
     digest: str
 
 
-class FileContentReadReq(BaseModel):
+class FileContentReadReq(BaseModelStrict):
     project_id: str
     digest: str
 
 
+class FilesStatsReq(BaseModelStrict):
+    project_id: str
+
+
 class FileContentReadRes(BaseModel):
     content: bytes
 
 
+class FilesStatsRes(BaseModel):
+    total_size_bytes: int
+
+
 class EnsureProjectExistsRes(BaseModel):
     project_name: str
 
 
-class CostCreateInput(BaseModel):
+class CostCreateInput(BaseModelStrict):
     prompt_token_cost: float
     completion_token_cost: float
     prompt_token_cost_unit: Optional[str] = Field(
@@ -802,7 +1011,7 @@ class CostCreateInput(BaseModel):
     )
 
 
-class CostCreateReq(BaseModel):
+class CostCreateReq(BaseModelStrict):
     project_id: str = Field(examples=["entity/project"])
     costs: dict[str, CostCreateInput]
     wb_user_id: Optional[str] = Field(None, description=WB_USER_ID_DESCRIPTION)
@@ -813,7 +1022,7 @@ class CostCreateRes(BaseModel):
     ids: list[tuple[str, str]]
 
 
-class CostQueryReq(BaseModel):
+class CostQueryReq(BaseModelStrict):
     project_id: str = Field(examples=["entity/project"])
     fields: Optional[list[str]] = Field(
         default=None,
@@ -856,7 +1065,7 @@ class CostQueryRes(BaseModel):
     results: list[CostQueryOutput]
 
 
-class CostPurgeReq(BaseModel):
+class CostPurgeReq(BaseModelStrict):
     project_id: str = Field(examples=["entity/project"])
     query: Query
 
@@ -865,7 +1074,7 @@ class CostPurgeRes(BaseModel):
     pass
 
 
-class ActionsExecuteBatchReq(BaseModel):
+class ActionsExecuteBatchReq(BaseModelStrict):
     project_id: str
     action_ref: str
     call_ids: list[str]
@@ -876,12 +1085,149 @@ class ActionsExecuteBatchRes(BaseModel):
     pass
 
 
+class ProjectStatsReq(BaseModelStrict):
+    project_id: str
+    include_trace_storage_size: Optional[bool] = True
+    include_object_storage_size: Optional[bool] = True
+    include_table_storage_size: Optional[bool] = True
+    include_file_storage_size: Optional[bool] = True
+
+
+class ProjectStatsRes(BaseModel):
+    trace_storage_size_bytes: int
+    objects_storage_size_bytes: int
+    tables_storage_size_bytes: int
+    files_storage_size_bytes: int
+
+
+# Thread API
+
+
+class ThreadSchema(BaseModel):
+    thread_id: str
+    turn_count: int = Field(description="Number of turn calls in this thread")
+    start_time: datetime.datetime = Field(
+        description="Earliest start time of turn calls in this thread"
+    )
+    last_updated: datetime.datetime = Field(
+        description="Latest end time of turn calls in this thread"
+    )
+    first_turn_id: Optional[str] = Field(
+        description="Turn ID of the first turn in this thread (earliest start_time)"
+    )
+    last_turn_id: Optional[str] = Field(
+        description="Turn ID of the latest turn in this thread (latest end_time)"
+    )
+    p50_turn_duration_ms: Optional[float] = Field(
+        description="50th percentile (median) of turn durations in milliseconds within this thread"
+    )
+    p99_turn_duration_ms: Optional[float] = Field(
+        description="99th percentile of turn durations in milliseconds within this thread"
+    )
+
+
+class ThreadsQueryFilter(BaseModelStrict):
+    after_datetime: Optional[datetime.datetime] = Field(
+        default=None,
+        description="Only include threads with start_time after this timestamp",
+        examples=["2024-01-01T00:00:00Z"],
+    )
+    before_datetime: Optional[datetime.datetime] = Field(
+        default=None,
+        description="Only include threads with last_updated before this timestamp",
+        examples=["2024-12-31T23:59:59Z"],
+    )
+    thread_ids: Optional[list[str]] = Field(
+        default=None,
+        description="Only include threads with thread_ids in this list",
+        examples=[["thread_1", "thread_2", "my_thread_id"]],
+    )
+
+
+class ThreadsQueryReq(BaseModelStrict):
+    """
+    Query threads with aggregated statistics based on turn calls only.
+
+    Turn calls are the immediate children of thread contexts (where call.id == turn_id).
+    This provides meaningful conversation-level statistics rather than including all
+    nested implementation details.
+    """
+
+    project_id: str = Field(
+        description="The ID of the project", examples=["my_entity/my_project"]
+    )
+    filter: Optional[ThreadsQueryFilter] = Field(
+        default=None,
+        description="Filter criteria for the threads query",
+    )
+    limit: Optional[int] = Field(
+        default=None, description="Maximum number of threads to return"
+    )
+    offset: Optional[int] = Field(default=None, description="Number of threads to skip")
+    sort_by: Optional[list[SortBy]] = Field(
+        default=None,
+        description="Sorting criteria for the threads. Supported fields: 'thread_id', 'turn_count', 'start_time', 'last_updated', 'p50_turn_duration_ms', 'p99_turn_duration_ms'.",
+        examples=[[SortBy(field="last_updated", direction="desc")]],
+    )
+
+
+class EvaluateModelReq(BaseModelStrict):
+    project_id: str
+    evaluation_ref: str
+    model_ref: str
+    wb_user_id: Optional[str] = Field(None, description=WB_USER_ID_DESCRIPTION)
+    # Fixes the following warning:
+    # UserWarning: Field "model_ref" has conflict with protected namespace "model_".
+    model_config = ConfigDict(protected_namespaces=())
+
+
+class EvaluateModelRes(BaseModel):
+    call_id: str
+
+
+class EvaluationStatusReq(BaseModelStrict):
+    project_id: str
+    call_id: str
+
+
+class EvaluationStatusNotFound(BaseModelStrict):
+    code: Literal["not_found"] = "not_found"
+
+
+class EvaluationStatusRunning(BaseModelStrict):
+    code: Literal["running"] = "running"
+    completed_rows: int
+    total_rows: int
+
+
+class EvaluationStatusFailed(BaseModelStrict):
+    code: Literal["failed"] = "failed"
+    error: Optional[str] = None
+
+
+class EvaluationStatusComplete(BaseModelStrict):
+    code: Literal["complete"] = "complete"
+    output: dict[str, Any]
+
+
+class EvaluationStatusRes(BaseModel):
+    status: Union[
+        EvaluationStatusNotFound,
+        EvaluationStatusRunning,
+        EvaluationStatusFailed,
+        EvaluationStatusComplete,
+    ]
+
+
 class TraceServerInterface(Protocol):
     def ensure_project_exists(
         self, entity: str, project: str
     ) -> EnsureProjectExistsRes:
         return EnsureProjectExistsRes(project_name=project)
 
+    # OTEL API
+    def otel_export(self, req: OtelExportReq) -> OtelExportRes: ...
+
     # Call API
     def call_start(self, req: CallStartReq) -> CallStartRes: ...
     def call_end(self, req: CallEndReq) -> CallEndRes: ...
@@ -891,6 +1237,7 @@ class TraceServerInterface(Protocol):
     def calls_delete(self, req: CallsDeleteReq) -> CallsDeleteRes: ...
     def calls_query_stats(self, req: CallsQueryStatsReq) -> CallsQueryStatsRes: ...
     def call_update(self, req: CallUpdateReq) -> CallUpdateRes: ...
+    def call_start_batch(self, req: CallCreateBatchReq) -> CallCreateBatchRes: ...
 
     # Op API
     def op_create(self, req: OpCreateReq) -> OpCreateRes: ...
@@ -906,15 +1253,36 @@ class TraceServerInterface(Protocol):
     def obj_create(self, req: ObjCreateReq) -> ObjCreateRes: ...
     def obj_read(self, req: ObjReadReq) -> ObjReadRes: ...
     def objs_query(self, req: ObjQueryReq) -> ObjQueryRes: ...
+    def obj_delete(self, req: ObjDeleteReq) -> ObjDeleteRes: ...
+
+    # Table API
     def table_create(self, req: TableCreateReq) -> TableCreateRes: ...
+    def table_create_from_digests(
+        self, req: TableCreateFromDigestsReq
+    ) -> TableCreateFromDigestsRes: ...
+
     def table_update(self, req: TableUpdateReq) -> TableUpdateRes: ...
     def table_query(self, req: TableQueryReq) -> TableQueryRes: ...
     def table_query_stream(self, req: TableQueryReq) -> Iterator[TableRowSchema]: ...
     def table_query_stats(self, req: TableQueryStatsReq) -> TableQueryStatsRes: ...
+    def table_query_stats_batch(
+        self, req: TableQueryStatsBatchReq
+    ) -> TableQueryStatsBatchRes: ...
+
+    # Ref API
     def refs_read_batch(self, req: RefsReadBatchReq) -> RefsReadBatchRes: ...
+
+    # File API
     def file_create(self, req: FileCreateReq) -> FileCreateRes: ...
     def file_content_read(self, req: FileContentReadReq) -> FileContentReadRes: ...
+    def files_stats(self, req: FilesStatsReq) -> FilesStatsRes: ...
+
+    # Feedback API
     def feedback_create(self, req: FeedbackCreateReq) -> FeedbackCreateRes: ...
+    def feedback_create_batch(
+        self, req: FeedbackCreateBatchReq
+    ) -> FeedbackCreateBatchRes: ...
+
     def feedback_query(self, req: FeedbackQueryReq) -> FeedbackQueryRes: ...
     def feedback_purge(self, req: FeedbackPurgeReq) -> FeedbackPurgeRes: ...
     def feedback_replace(self, req: FeedbackReplaceReq) -> FeedbackReplaceRes: ...
@@ -926,3 +1294,21 @@ class TraceServerInterface(Protocol):
 
     # Execute LLM API
     def completions_create(self, req: CompletionsCreateReq) -> CompletionsCreateRes: ...
+
+    # Execute LLM API (Streaming)
+    # Returns an iterator of JSON-serializable chunks that together form the streamed
+    # response from the model provider. Each element must be a dictionary that can
+    # be serialized with ``json.dumps``.
+    def completions_create_stream(
+        self, req: CompletionsCreateReq
+    ) -> Iterator[dict[str, Any]]: ...
+
+    # Project statistics API
+    def project_stats(self, req: ProjectStatsReq) -> ProjectStatsRes: ...
+
+    # Thread API
+    def threads_query_stream(self, req: ThreadsQueryReq) -> Iterator[ThreadSchema]: ...
+
+    # Evaluation API
+    def evaluate_model(self, req: EvaluateModelReq) -> EvaluateModelRes: ...
+    def evaluation_status(self, req: EvaluationStatusReq) -> EvaluationStatusRes: ...
